""" Defines soil.data()"""
from typing import (
    Any,
    Dict,
    Optional,
    Type,
    TypeVar,
    TypedDict,
    cast,
    overload,
    TypeAlias,
)

from pandas import DataFrame
from numpy import ndarray

from soil.api import get_result, upload_data, get_alias
from soil.data_structure import DataStructure, get_data_structure_name_and_serialize
from soil import errors
from soil.types import SerializableDataStructure


_GenericDataStructure = TypeVar(
    "_GenericDataStructure", bound=SerializableDataStructure
)
_TypedDict = TypeVar("_TypedDict", bound=TypedDict)
_DataObject: TypeAlias = dict | list | ndarray | DataFrame | _TypedDict


@overload
def data(
    data_object: str,
    metadata: None = None,
    *,
    return_type: None = None,
) -> DataStructure:
    ...


@overload
def data(
    data_object: str,
    metadata: None = None,
    *,
    return_type: Type[_GenericDataStructure],
) -> _GenericDataStructure:
    ...


@overload
def data(
    data_object: _DataObject,
    metadata: dict[str, Any] | None = None,
    *,
    return_type: None = None,
) -> DataStructure:
    ...


@overload
def data(
    data_object: _DataObject,
    metadata: dict[str, Any] | None = None,
    *,
    return_type: Type[_GenericDataStructure],
) -> _GenericDataStructure:
    ...


def data(
    data_object: str | _DataObject,
    metadata: dict[str, Any] | None = None,
    *,
    return_type: Type[_GenericDataStructure] | None = None,
) -> _GenericDataStructure | DataStructure:
    """Load data from the cloud or mark it as uploadable"""
    cast_return_type = DataStructure if return_type is None else return_type
    if isinstance(data_object, str):
        # Data object is an id or an alias
        try:
            data_object = _load_data_alias(data_object)
        except errors.DataNotFound:
            pass
        return cast(cast_return_type, _load_data_id(data_object))
    return cast(cast_return_type, _upload_data(data_object, metadata))


def _upload_data(
    data_object: Any, metadata: Optional[Dict[str, Any]] = None
) -> DataStructure:
    ds_name, serialized = get_data_structure_name_and_serialize(data_object)
    result = upload_data(ds_name, serialized, metadata)
    ds = DataStructure(result["_id"], dstype=result["type"])
    return ds


def _load_datastructure(did: str, dtype: str) -> DataStructure:
    # TODO: dynamically load a data structure
    return DataStructure(did, dstype=dtype)


def _load_data_alias(alias: str) -> str:
    return get_alias(alias)["state"]["result_id"]


def _load_data_id(data_id: str) -> DataStructure:
    result = get_result(data_id)
    return _load_datastructure(result["_id"], result["type"])



""" Exports the soil logger """
import logging
import logging.handlers
from typing import cast, LiteralString
from enum import StrEnum
from os.path import exists
import sqlite3

_SOIL_LOGGER = "soil_logger"
PROCESSED_FILES_DB = "/rain-data/processed_files.db"

logger = logging.getLogger(_SOIL_LOGGER)  # pylint: disable=invalid-name


class TypeLog(StrEnum):
    """Types of Logs."""

    PROCESSED = "processed"
    NOT_CONSISTENT = "not_consistent"
    NOT_PROCESSED = "not_processed"


class HashFileHandler(logging.StreamHandler):
    """Handler of logs to store file to sqlite."""

    type_log = "type"
    hashfile = "file"
    state_name_column = "state_description"
    message_column = "message"

    def emit(self, record: logging.LogRecord) -> None:
        type_log = cast(TypeLog | None, getattr(record, "type", None))
        hashfile = cast(str | None, getattr(record, "file", None))
        if (
            type_log is not None
            and hashfile is not None
            and type_log in [*TypeLog]
            and exists(PROCESSED_FILES_DB)
        ):
            self._process_file_in_storage(
                hashfile=hashfile, type_log=type_log, message=record.getMessage()
            )

    def _process_file_in_storage(self, **kwargs) -> None:
        """Process the file hash to update the DB."""
        with sqlite3.connect(PROCESSED_FILES_DB) as connection:
            cursor = connection.cursor()
            # pylint:disable=use-maxsplit-arg
            table_name = PROCESSED_FILES_DB.split("/")[-1].split(".", maxsplit=1)[0]
            self._alter_table_if_necessary(cursor=cursor, table_name=table_name)
            self._update_hashfile(cursor=cursor, table_name=table_name, **kwargs)

    def _update_hashfile(
        self,
        *,
        cursor: sqlite3.Cursor,
        table_name: LiteralString,
        hashfile: str,
        type_log: TypeLog,
        message: str,
    ) -> None:
        cursor.execute(
            f"UPDATE {table_name} "
            f"SET {self.state_name_column} = ?, {self.message_column} = ? WHERE file_name = ?",
            (type_log, message, hashfile),
        )

    def _alter_table_if_necessary(
        self, *, cursor: sqlite3.Cursor, table_name: LiteralString
    ) -> None:
        pragmas = cursor.execute(f"PRAGMA table_info({table_name});")
        columns: list[str] = [n for _, n, *_ in pragmas.fetchall()]
        if self.state_name_column not in columns:
            # sqlite current version does not support adding checks on existing tables.
            cursor.execute(
                f"ALTER TABLE {table_name}"
                f"\n\tADD COLUMN {self.state_name_column} VARCHAR;"
            )
        if self.message_column not in columns:
            cursor.execute(
                f"ALTER TABLE {table_name}"
                f"\n\tADD COLUMN {self.message_column} VARCHAR;"
            )


logger.addHandler(HashFileHandler())



"""This module implements SOIL's Command LIne Interface"""
import argparse
import sys
import os
import json
import shutil
import getpass
from typing import Dict, Tuple, Optional
from json import JSONDecodeError
import subprocess  # nosec
import requests
import yaml

parser = argparse.ArgumentParser(prog="soil")
subparsers = parser.add_subparsers(dest="command")

login_cmd = subparsers.add_parser("login", help="Login to configured soil instance")
login_cmd.add_argument("--user", type=str, help="User name or email of the user")
login_cmd.add_argument("--password", type=str, help="Password")


run_cmd = subparsers.add_parser(
    "run", help="Runs the provided module in the current soil project"
)
run_cmd.add_argument(
    "chapter",
    metavar="chapter",
    help="Chapter in soil.yml (setup, data, migration, ...)",
)
run_cmd.add_argument("module", metavar="module", help="Module of chapter in soil.yml")
run_cmd.add_argument(
    "module_args", nargs=argparse.REMAINDER, help="Arguments to be passed to the module"
)


def _save_environment(config: Dict, env: Dict) -> None:
    """
    Saves the environment keys to a file.
    The file is located in ~/.soil/soil.env and ~/.soil/:app_id/soil.env
    Where app_id is the app_id in the config.
    Notice that ~/.soil/soil.env will store the last login only.
    """

    try:
        os.makedirs(os.path.expanduser("~/.soil/" + config["auth_app_id"]))
    except FileExistsError:
        # directory already exists
        pass

    with open(
        os.path.expanduser("~/.soil/soil.env"), "w", encoding="utf-8"
    ) as env_file:
        env_file.write(json.dumps(env, indent=4, sort_keys=True))

    shutil.copyfile(
        os.path.expanduser("~/.soil/soil.env"),
        os.path.expanduser("~/.soil/" + config["auth_app_id"] + "/soil.env"),
    )


def _get_soil_root(relpath: str):  # type: ignore
    """Checks if the current dir is under a soil environment
    and returns its root. Returns None otherwise."""

    path = os.path.abspath(relpath) + "/"

    while path != "/":
        path, _ = os.path.split(path)
        if "soil.yml" in os.listdir(path):
            return path

    return None


def _soil_init() -> Tuple[Dict, Optional[Dict]]:
    """Loads configuration and environment."""

    project_root = _get_soil_root(".")
    try:
        if project_root is None:
            raise FileNotFoundError("Project root not found.")
        with open(project_root + "/soil.conf", encoding="utf-8") as config_file:
            config = json.loads(config_file.read())
    except FileNotFoundError:
        try:
            with open(
                os.path.expanduser("~/.soil/soil.conf"), "r", encoding="utf-8"
            ) as config_file:
                config = json.loads(config_file.read())
        except (IOError, JSONDecodeError):
            if sys.argv[1] != "configure":
                try:
                    os.rename(
                        os.path.expanduser("~/.soil/soil.conf.bak"),
                        os.path.expanduser("~/.soil/soil.conf"),
                    )
                except IOError:
                    print(
                        "Can not load soil configuration. Plase run soil configure to configure it."
                    )
                    sys.exit(1)
            else:
                config = None
    try:
        with open(
            os.path.expanduser("~/.soil/soil.env"), "r", encoding="utf-8"
        ) as env_file:
            env = json.loads(env_file.read())
        return config, env
    except (IOError, JSONDecodeError):
        if sys.argv[1] != "configure" and sys.argv[1] != "login":
            try:
                os.rename(
                    os.path.expanduser("~/.soil/soil.conf.bak"),
                    os.path.expanduser("~/.soil/soil.conf"),
                )
            except IOError:
                print(
                    "Can not load soil environment. Plase run soil configure to initialize it."
                )
                sys.exit(1)
        return config, None


def _login(args: argparse.Namespace, config: Dict, env: Optional[Dict]) -> None:
    """
    Authenticates to the authentication backend and stores the credentials (JWT) in the environment
    """

    print(f"Authenticating to {config['auth_url']} for app {config['auth_app_id']}...")

    if env is None:
        env = {}

    if args.user is not None and args.password is not None:
        username = args.user
        password = args.password
    else:
        if "auth" in env:
            username = input(
                "Username: ["
                + env["auth"]["user"].get(
                    "username", env["auth"]["user"].get("email", "")
                )
                + "]"
            )  # nosec - Input is safe in python3
            if username == "":
                username = env["auth"]["user"].get(
                    "username", env["auth"]["user"].get("email", "")
                )
        else:
            username = input("Username: ")  # nosec - Input is safe in python3
        password = getpass.getpass()

    request_json = {
        "loginId": username,
        "password": password,
        "applicationId": config["auth_app_id"],
    }

    resp = requests.post(
        config["auth_url"] + "/api/login",
        headers={"Authorization": config["auth_api_key"]},
        json=request_json,
        timeout=10,
    )

    if resp.status_code in {200, 202}:
        env["auth"] = json.loads(resp.content)
        _save_environment(config, env)
        print(
            "Successfully logged in as "
            + env["auth"]["user"].get("username", env["auth"]["user"].get("email", ""))
            + "!"
        )
    elif resp.status_code == 404:
        print("The user was not found or the password was incorrect.")
        sys.exit(1)
    elif resp.status_code == 400:
        print(
            f"Invalid parameters. Perhaps the api key {config['auth_api_key']} or \
                 the app id {config['auth_app_id']} are not correct."
        )
        sys.exit(1)
    else:
        print("Login failed with status code", resp.status_code)
        sys.exit(1)


def _run(args: argparse.Namespace) -> None:
    """
    Runs the script provided as argument using the python in the virtual environment
    """

    soil_root = _get_soil_root(".")
    if not soil_root:
        print(
            "This folder is not initalized as a soil project. "
            "Please run soil init to initialize it."
        )
        sys.exit(1)

    chapter = vars(args)["chapter"]
    module = vars(args)["module"]
    module_args = vars(args)["module_args"]

    os.chdir(soil_root)
    with open("soil.yml", encoding="utf-8") as f:
        conf = yaml.safe_load(f)
    try:
        script = [script for script in conf[chapter] if script.get(module) is not None][
            0
        ][module]
    except (IndexError, KeyError):
        print("Script", module, "not found.")
        sys.exit(1)
    params = script.get("params", {})
    params = [["--" + k, str(v)] for (k, v) in params.items()]
    params = [item for sublist in params for item in sublist]
    try:
        shell = [".venv/bin/python", "-m", script["path"]] + module_args + params
        print("Running:", " ".join(shell))
        subprocess.run(shell, check=True)  # nosec
    except subprocess.CalledProcessError:
        sys.exit(1)


def main() -> None:
    """main function"""

    args = parser.parse_args()

    config, env = _soil_init()

    try:
        if args.command == "login":
            _login(args, config, env)
        elif args.command == "run":
            _run(args)
    except KeyboardInterrupt:
        print("\nKeyboardInterrupt - Exit...")
        try:
            sys.exit(0)
        except SystemExit:
            os._exit(0)  # NOQA # pylint: disable=protected-access


if __name__ == "__main__":
    main()



# pylint:disable=missing-module-docstring,missing-class-docstring,missing-function-docstring
# pylint:disable=import-outside-toplevel
import os
import unittest
from unittest.mock import patch
import logging
import sqlite3

_TEST_LOGGER = "test_logger"
PROCESSED_FILES_DB = "test/processed_files.db"


@patch("soil.logger.PROCESSED_FILES_DB", new=PROCESSED_FILES_DB)
@patch("soil.logger._SOIL_LOGGER", new=_TEST_LOGGER)
class TestHashfileLogger(unittest.TestCase):
    @staticmethod
    def _init_storage():
        """Checks if the state files are writeable."""
        if not os.path.exists(PROCESSED_FILES_DB):
            os.mknod(PROCESSED_FILES_DB)
        with sqlite3.connect(PROCESSED_FILES_DB) as connection:
            connection.execute("DROP TABLE IF EXISTS processed_files")
            connection.execute(
                """
                CREATE TABLE processed_files (
                    file_name VARCHAR PRIMARY KEY,
                    date VARCHAR NOT NULL,
                    size int NOT NULL,
                    datatype VARCHAR NOT NULL,
                    state VARCHAR CHECK( state IN ('WAITING','PROCESSED','ERROR') ) NOT NULL
                )
                """
            )

    @staticmethod
    def _store_file_db(hexdigest, date, size, datatype):
        with sqlite3.connect(PROCESSED_FILES_DB) as connection:
            connection.execute(
                """
                INSERT INTO processed_files (file_name, date, size, datatype, state)
                VALUES (?,?,?,?,?)
                """,
                (hexdigest, date, size, datatype, "WAITING"),
            )

    def test_logs_handler(self):
        # importing inside method because of patch test_logger
        from soil.logger import logger

        TestHashfileLogger._init_storage()
        file = "hash_file_1"
        TestHashfileLogger._store_file_db(file, "2022-02-02", 1, "test1")
        logger.setLevel(logging.DEBUG)
        logger.debug(
            "%s has been processed successfully!",
            file,
            extra={"type": "processed", "file": file},
        )
        logger.debug(
            "%s has been processed successfully!",
            "file_not_existent",
            extra={"type": "processed", "file": "file_not_existent"},
        )
        with sqlite3.connect(PROCESSED_FILES_DB) as connection:
            results = connection.execute(
                """
                    SELECT *
                    FROM processed_files
                """
            ).fetchall()
        self.assertListEqual(
            [
                (
                    "hash_file_1",
                    "2022-02-02",
                    1,
                    "test1",
                    "WAITING",
                    "processed",
                    f"{file} has been processed successfully!",
                )
            ],
            results,
        )



""" Module for BaseStorage """
from dataclasses import dataclass
from typing import Type, TypeVar, Any

StorageClass = TypeVar("StorageClass")


@dataclass
class BaseStorage:
    """
    Abstract class that implements serialize and deserialize methods for storage classes.
    """

    def serialize(self) -> dict[str, Any]:
        """
        Serializes the storage object. In general it shouldn't be used from a module.
        """

    @classmethod
    def deserialize(
        cls: Type[StorageClass], serialized_storage_object: dict[str, Any]
    ) -> StorageClass:
        """
        Takes a serialized storage object and returns an instance.
        """



""" Defines the configuration for Soil """
import json
import logging
import os
import time
from os import getenv
from typing import NamedTuple, Optional

import jwt
import requests

from soil import errors

# TODO add windows support

# pylint: disable=invalid-name
env = getenv("PY_ENV", "development")


def _refresh_token(auth_host: str, refresh_token: str) -> str:
    url = auth_host + "/api/jwt/refresh"
    response = requests.post(url, json={"refreshToken": refresh_token}, timeout=30)
    if response.status_code != 200:
        raise errors.LoginError("Invalid refresh token please run soil login again.")
    return json.loads(response.text)["token"]


def get_soil_root(relpath: str) -> Optional[str]:
    """Checks if the current dir is under a soil environment
    and returns its root. Returns None otherwise."""
    path = os.path.abspath(relpath) + "/"
    while path != "/":
        path, _ = os.path.split(path)
        if "soil.yml" in os.listdir(path):
            return path
    return None


project_root = get_soil_root(".")
SOIL_URL = ""
if env != "test":
    try:
        if project_root is None:
            raise FileNotFoundError("Project root not found.")
        with open(project_root + "/soil.conf", encoding="utf-8") as conf_file:
            CONF = json.loads(conf_file.read())
            SOIL_URL = CONF["soil_url"]
        logging.info(
            "Loaded config from %s using app_id %s",
            project_root + "/soil.conf",
            CONF["auth_app_id"],
        )
    except FileNotFoundError:
        try:
            with open(
                getenv("HOME", "") + "/.soil/soil.conf", encoding="utf-8"
            ) as conf_file:
                CONF = json.loads(conf_file.read())
                SOIL_URL = CONF["soil_url"]
            logging.warning(
                "Loaded config from %s using app_id %s. It is recommended to have a soil.conf file at the root of the project.",  # NOQA pylint:disable=line-too-long
                getenv("HOME", "") + "/.soil/soil.conf",
                CONF["auth_app_id"],
            )

        except FileNotFoundError:
            logging.error(
                "~/.soil/soil.conf file not found. Please run soil configure."
            )
else:
    SOIL_URL = "http://test_host.test"

TOKEN = ""  # nosec

if env != "test":
    try:
        with open(
            getenv("HOME", "") + "/.soil/" + CONF["auth_app_id"] + "/soil.env",
            encoding="utf-8",
        ) as env_file:
            ENV = json.loads(env_file.read())
            TOKEN = _refresh_token(CONF["auth_url"], ENV["auth"]["refreshToken"])
    except FileNotFoundError:
        logging.warning(
            (
                "~/.soil/%s/soil.env file not found."
                " Using ~/.soil/soil.env instead. It is recommended to run soil login."
            ),
            CONF["auth_app_id"],
        )
        try:
            with open(
                getenv("HOME", "") + "/.soil/soil.env", encoding="utf-8"
            ) as env_file:
                ENV = json.loads(env_file.read())
                TOKEN = _refresh_token(CONF["auth_url"], ENV["auth"]["refreshToken"])
        except FileNotFoundError:
            logging.error("~/.soil/soil.env file not found. Please run soil login.")
else:
    TOKEN = "test_token"  # nosec

DEFAULT_CONFIG = {
    "host": getenv("SOIL_HOST", SOIL_URL),
    "token": getenv("SOIL_TOKEN", TOKEN),
}

if env != "test":
    decoded_token = jwt.decode(TOKEN, options={"verify_signature": False})
    conf_app_id = CONF["auth_app_id"].strip()
    token_app_id = decoded_token["applicationId"]

    if token_app_id != conf_app_id:
        raise errors.LoginError(
            "Application Id for the project config "
            + f"{conf_app_id} does not coincide with the token application id {token_app_id}.\n"
            + "Please, run 'soil login' again."
        )


class SoilConfiguration(NamedTuple):
    """Soil configuration class"""

    host: Optional[str]
    token: Optional[str]

    def __getattribute__(self, name: str) -> str:
        """Refresh the token if necessary before accessing it."""
        # pylint:disable=global-statement
        global TOKEN

        if name != "token":
            return object.__getattribute__(self, name)

        if env == "test":
            return TOKEN

        # Min nummber of seconds before expire that will trigger a refresh.
        min_delta = 10

        exp = jwt.decode(TOKEN, options={"verify_signature": False})["exp"]
        if exp - time.time() < min_delta:
            TOKEN = _refresh_token(CONF["auth_url"], ENV["auth"]["refreshToken"])
        return TOKEN


GLOBAL_CONFIG = SoilConfiguration(**DEFAULT_CONFIG)


# Not used for now
# def config(token: Optional[str] = None, host: Optional[str] = None) -> SoilConfiguration:
#     ''' Set the Soil's configuration '''
#     global GLOBAL_CONFIG  # pylint: disable=global-statement
#     new_config = SoilConfiguration(token=token, host=host)
#     GLOBAL_CONFIG = SoilConfiguration(**{**GLOBAL_CONFIG._asdict(), **new_config._asdict()})
#     return GLOBAL_CONFIG



# pylint: disable=missing-docstring,line-too-long,unused-argument,no-member,too-many-locals
# type: ignore

import unittest
from unittest.mock import patch
import soil
from soil.data_structure import DataStructure
from soil import modulify


class List:
    # pylint:disable=too-few-public-methods
    pass


@modulify(output_types=lambda *input_types, **args: [List])
def merge_list(data1, data2):
    pass


@modulify()
def merge_list_with_signature(data1, data2) -> tuple[DataStructure]:
    pass


@modulify()
def merge_list_with_signature_error(data1, data2) -> DataStructure:
    pass


@modulify
def merge_list_with_signature_error_2(data1, data2):
    pass


@modulify(output_types=lambda *input_types, **args: [List, List])
def split_list(data):
    pass


def soil_data_patch(name):
    return DataStructure(name)


@patch("soil.data", side_effect=soil_data_patch)
@patch(
    "soil.utils.uuid4",
    side_effect=list(range(1111, 1111 * 1000, 1111)),
)
@patch("soil.api.create_experiment")
@patch("soil.api.get_experiment")
@patch("soil.api.get_experiment_logs")
@patch("soil.api.get_result_data")
class TestPipelines(unittest.TestCase):
    # pylint:disable=too-many-function-args

    @patch("soil.api.get_result")
    def test_pipeline_simple(
        self,
        get_result_mock,
        get_result_data_mock,
        experiment_logs_mock,
        get_experiment_mock,
        create_experiment_mock,
        random_int_patch,
        mock_data_patch,
    ):
        data_ref_1 = soil.data("data_ref_1")
        data_ref_2 = soil.data("data_ref_2")
        (result,) = merge_list(data_ref_1, data_ref_2)
        get_experiment_mock.side_effect = [
            {"experiment_status": "EXECUTING"},
            {"experiment_status": "DONE"},
        ]
        experiment_logs_mock.return_value = []

        _data = result.metadata  # NOQA

        assert len(experiment_logs_mock.call_args_list) == 2
        assert len(get_result_data_mock.call_args_list) == 0
        assert len(get_result_mock.call_args_list) == 1
        assert len(get_experiment_mock.call_args_list) == 2
        create_experiment_mock.assert_called_once_with(
            [
                {
                    "id": "unit.soil.test_pipelines.merge_list-2222",
                    "module": "unit.soil.test_pipelines.merge_list",
                    "inputs": ["data_ref_1", "data_ref_2"],
                    "outputs": ["unit/soil/test_pipelines/merge_list-1111-0"],
                    "args": {},
                }
            ]
        )

    @patch("soil.api.get_result")
    def test_pipeline_simple_with_signature(
        self,
        get_result_mock,
        get_result_data_mock,
        experiment_logs_mock,
        get_experiment_mock,
        create_experiment_mock,
        *_unused_mocks,
    ):
        data_ref_1 = soil.data("data_ref_1")
        data_ref_2 = soil.data("data_ref_2")
        (result,) = merge_list_with_signature(data_ref_1, data_ref_2)
        get_experiment_mock.side_effect = [
            {"experiment_status": "EXECUTING"},
            {"experiment_status": "DONE"},
        ]
        experiment_logs_mock.return_value = []

        _data = result.metadata  # NOQA

        assert len(experiment_logs_mock.call_args_list) == 2
        assert len(get_result_data_mock.call_args_list) == 0
        assert len(get_result_mock.call_args_list) == 1
        assert len(get_experiment_mock.call_args_list) == 2
        create_experiment_mock.assert_called_once_with(
            [
                {
                    "id": "unit.soil.test_pipelines.merge_list_with_signature-2222",
                    "module": "unit.soil.test_pipelines.merge_list_with_signature",
                    "inputs": ["data_ref_1", "data_ref_2"],
                    "outputs": [
                        "unit/soil/test_pipelines/merge_list_with_signature-1111-0"
                    ],
                    "args": {},
                }
            ]
        )

    @patch("soil.api.get_result")
    def test_pipeline_simple_with_signature_raise_error(
        self,
        *_unused_mocks,
    ):
        data_ref_1 = soil.data("data_ref_1")
        data_ref_2 = soil.data("data_ref_2")
        with self.assertRaises(ValueError):
            merge_list_with_signature_error(data_ref_1, data_ref_2)

        with self.assertRaises(ValueError):
            merge_list_with_signature_error_2(data_ref_1, data_ref_2)

    def test_pipeline_split(
        self,
        get_result_data_mock,
        experiment_logs_mock,
        get_experiment_mock,
        create_experiment_mock,
        random_int_patch,
        mock_data_patch,
    ):
        data_ref_1 = soil.data("data_ref_1")
        data_ref_2 = soil.data("data_ref_2")
        (result1, result2) = split_list(data_ref_1, data_ref_2)
        get_experiment_mock.side_effect = [{"experiment_status": "DONE"}]
        experiment_logs_mock.return_value = []

        _data = result1.data  # NOQA
        _data = result2.data  # NOQA

        assert len(experiment_logs_mock.call_args_list) == 1
        assert len(get_result_data_mock.call_args_list) == 2
        assert len(get_experiment_mock.call_args_list) == 1

        create_experiment_mock.assert_called_once_with(
            [
                {
                    "id": "unit.soil.test_pipelines.split_list-2222",
                    "module": "unit.soil.test_pipelines.split_list",
                    "inputs": ["data_ref_1", "data_ref_2"],
                    "outputs": [
                        "unit/soil/test_pipelines/split_list-1111-0",
                        "unit/soil/test_pipelines/split_list-1111-1",
                    ],
                    "args": {},
                }
            ]
        )

    def test_pipeline_split_merge(
        self,
        get_result_data_mock,
        experiment_logs_mock,
        get_experiment_mock,
        create_experiment_mock,
        random_int_patch,
        mock_data_patch,
    ):
        data_ref_1 = soil.data("data_ref_1")
        (res_1, res_2) = split_list(data_ref_1)
        (result1,) = merge_list(res_1, res_2)
        get_experiment_mock.side_effect = [{"experiment_status": "DONE"}]
        experiment_logs_mock.return_value = []

        _data = result1.get_data()  # NOQA

        assert len(experiment_logs_mock.call_args_list) == 1
        assert len(get_result_data_mock.call_args_list) == 1
        assert len(get_experiment_mock.call_args_list) == 1

        create_experiment_mock.assert_called_once_with(
            [
                {
                    "id": "unit.soil.test_pipelines.split_list-2222",
                    "module": "unit.soil.test_pipelines.split_list",
                    "inputs": ["data_ref_1"],
                    "outputs": [
                        "unit/soil/test_pipelines/split_list-1111-0",
                        "unit/soil/test_pipelines/split_list-1111-1",
                    ],
                    "args": {},
                },
                {
                    "id": "unit.soil.test_pipelines.merge_list-4444",
                    "module": "unit.soil.test_pipelines.merge_list",
                    "inputs": [
                        "unit/soil/test_pipelines/split_list-1111-0",
                        "unit/soil/test_pipelines/split_list-1111-1",
                    ],
                    "outputs": ["unit/soil/test_pipelines/merge_list-3333-0"],
                    "args": {},
                },
            ]
        )

    def test_pipeline_complex_split_merge(
        self,
        get_result_data_mock,
        experiment_logs_mock,
        get_experiment_mock,
        create_experiment_mock,
        random_int_patch,
        mock_data_patch,
    ):
        data_ref_1 = soil.data("data_ref_1")
        (res_1, res_2) = split_list(data_ref_1)
        (res_11, res_12) = split_list(res_1)
        (res_21, res_22) = split_list(res_2)
        (result1,) = merge_list(res_11, res_21)
        (result2,) = merge_list(res_12, res_22)
        (result,) = merge_list(result1, result2)
        get_experiment_mock.side_effect = [{"experiment_status": "DONE"}]
        experiment_logs_mock.return_value = []

        _data = result.get_data()  # NOQA

        assert len(experiment_logs_mock.call_args_list) == 1
        assert len(get_result_data_mock.call_args_list) == 1
        assert len(get_experiment_mock.call_args_list) == 1

        create_experiment_mock.assert_called_once_with(
            [
                {
                    "id": "unit.soil.test_pipelines.merge_list-8888",
                    "module": "unit.soil.test_pipelines.merge_list",
                    "inputs": [
                        "unit/soil/test_pipelines/split_list-3333-0",
                        "unit/soil/test_pipelines/split_list-5555-0",
                    ],
                    "outputs": ["unit/soil/test_pipelines/merge_list-7777-0"],
                    "args": {},
                },
                {
                    "id": "unit.soil.test_pipelines.split_list-4444",
                    "module": "unit.soil.test_pipelines.split_list",
                    "inputs": ["unit/soil/test_pipelines/split_list-1111-0"],
                    "outputs": [
                        "unit/soil/test_pipelines/split_list-3333-0",
                        "unit/soil/test_pipelines/split_list-3333-1",
                    ],
                    "args": {},
                },
                {
                    "id": "unit.soil.test_pipelines.split_list-2222",
                    "module": "unit.soil.test_pipelines.split_list",
                    "inputs": ["data_ref_1"],
                    "outputs": [
                        "unit/soil/test_pipelines/split_list-1111-0",
                        "unit/soil/test_pipelines/split_list-1111-1",
                    ],
                    "args": {},
                },
                {
                    "id": "unit.soil.test_pipelines.split_list-6666",
                    "module": "unit.soil.test_pipelines.split_list",
                    "inputs": ["unit/soil/test_pipelines/split_list-1111-1"],
                    "outputs": [
                        "unit/soil/test_pipelines/split_list-5555-0",
                        "unit/soil/test_pipelines/split_list-5555-1",
                    ],
                    "args": {},
                },
                {
                    "id": "unit.soil.test_pipelines.merge_list-11110",
                    "module": "unit.soil.test_pipelines.merge_list",
                    "inputs": [
                        "unit/soil/test_pipelines/split_list-3333-1",
                        "unit/soil/test_pipelines/split_list-5555-1",
                    ],
                    "outputs": ["unit/soil/test_pipelines/merge_list-9999-0"],
                    "args": {},
                },
                {
                    "id": "unit.soil.test_pipelines.merge_list-13332",
                    "module": "unit.soil.test_pipelines.merge_list",
                    "inputs": [
                        "unit/soil/test_pipelines/merge_list-7777-0",
                        "unit/soil/test_pipelines/merge_list-9999-0",
                    ],
                    "outputs": ["unit/soil/test_pipelines/merge_list-12221-0"],
                    "args": {},
                },
            ]
        )
        # print(create_experiment_mock.call_args_list)
